import java.rmi.*;
import java.util.*;
import java.sql.*;

/** 
 *
 * This top level class defines the configuration of the test bed <br>
 * In the test bed, are the following entities. All sets are zero-based: <br>
 * <p>
 * Voters(V): Set of voters [nVOTERS]
 * Machines (M): Set of machines [nMACHINES] 
 * Voter->Machine Mapping. Every voter resides on exactly one machine
 * BLOBS (B): Set of BLOBS [nBLOBS]
 * <p>
 * A voter may be interested in one or more BLOBS
 * For every V, there exists a subset in B that he is interested in
 * Every BLOB is initially established on a machine M
 * For every BLOB in B, there exists exactly on entry in M, where the
 * BLOB resides.
 * <p>
 * The problem then, can be formulated thus:
 *
 * Do a Byzantine such that every member in B is replicated at a place 
 * that is most optimal for access by all voters requesting it.
 * Print the previous and the revised mapping of BLOBS to Machines
 * @Author: Preethi Vishwanath
 * @Date: 02/13/07
 */
public class Manager 
{
	/*
         * Please ensure that the number of voters
 	 * correspond exactly to the # of XML documents
	 * under the config/ directory
         * The program will exit otherwise
	 */

	private static final int nVOTERS = 11;
	private static final int nMACHINES = 8;
	private static final String mPROTOCOL = "rmi://";
	private static final String ipPREFIX = "192.168.0.";
	private static final int nBLOBS = 4;
	
	/* 
	 * A faulty voter changes his vote in every
	 * round. This will allow us to extend the program
	 * later
	 */

	private static int nFAULTY = 0;
	private static int faulty_voters[] = {2, 7};
	


	private static int maxDELAY = 100;

		
	/*
	 * A database of the same name exists
	 * on ALL machines participating in the
	 * distributed Byzantine agreement
	 */

	private static String dbName = "testdb";
	private static String dbUser = "testuser";
	private static String dbPassword = "testuser";
	
	/*
	 * The configuration table that sets up all
	 * the voters. This contains details about
	 * how each voter should vote for each blob
	 * This table needs to be present only
	 * in the manager.
	 */
 
	private static String tblConfig = "tblP";
	private static String tblMain = "tblMachByzantine";

	public Manager() {

	}
	/**
	 * Uses the VoterManager to create voters
	 * on different machines. Uses Java RMI to achieve the same.
	 * 
	 * @param vList: Output parameter populated with <br>
	 * 		 reference to the voter object <br>
	 * @param vIDtoM: Hashtable mapping the vID to VoterManager reference 
	 * @param vtr_mcID: Voter ID->Machine ID 
	 * @param blob_mcID: Initial Blob->Machine ID mapping 
	 * @param faulty_voters: Voters who are faulty, i.e. vote randomly <br>
	 * 			 every single time in a Byzatine process
	 * @param mcID_tIP: Machine->IP Address mapping
	 * @return none
	 * @Author: Preethi Vishwanath
	 */
	private static void createVoters(ArrayList vList, 
					 Hashtable vIDtoM,
					 int[] vtr_mcID,
					 int[] blob_mcID,
					 int[] faulty_voters,
					 String[] mcID_tIP) 
		throws ClassNotFoundException, SQLException, RemoteException
	{
		boolean isFaulty = false;
		int f_idx = 0;
		int delay = 0;
		int v_idx = 0;	
		Voter vtr = null;
		Connection db;
		Statement sql;
		ResultSet rs;
		VoterManager v;
		String sqlText;
		Random rnd = new Random();
	      	Class.forName("org.postgresql.Driver");
   		db = DriverManager.getConnection("jdbc:postgresql://localhost/" + dbName, 
						  dbUser, dbPassword);
      		sql = db.createStatement();
		for (v_idx = 0; v_idx < nVOTERS; v_idx++) {
			delay = rnd.nextInt(maxDELAY) + 1;
			isFaulty = false;
			if ((f_idx < nFAULTY) && 
			     (v_idx == faulty_voters[f_idx])) {
				isFaulty = true;
				f_idx ++;
			}
			v = (VoterManager)vIDtoM.get(new Integer(v_idx));	
			Hashtable ht_blobMachine = new Hashtable();
       			sqlText = "SELECT blob_id FROM " + tblConfig + " WHERE voter_id = " + v_idx;
      			rs = sql.executeQuery(sqlText);
      			while (rs.next()) {
				String mBlob = rs.getString(1);
				Integer m_Blob = new Integer(Integer.parseInt(mBlob));
				Integer m_Machine = new Integer(VoterConsts.UNDEFINED);
				ht_blobMachine.put(m_Blob, m_Machine);	
      			}
      			rs.close();
			vtr = v.createVoter(delay, 
					    isFaulty,
					    v_idx,
					    vtr_mcID[v_idx], 	
					    ht_blobMachine,
					    blob_mcID,
					    mcID_tIP,
					    dbName,
					    dbUser, 
					    dbPassword, 
					    tblMain);
			vList.add(vtr);
		}	
	}


	public static void main(String[] args)
	{		
		
		int[] blob_mcID = new int[nBLOBS];
		int[] blob_axsFreq = new int[nBLOBS];

		int[] vtr_mcID = new int[nVOTERS];
		String[] mcID_IP = new String[nMACHINES];

		VoterManager[] vMRef = new VoterManager[nMACHINES];
		Hashtable ht_VtoM = new Hashtable();
	
		Random rnd = new Random();	
		
		int v_idx, b_idx, m_idx;	
		VoterManager v;

		if (System.getSecurityManager() == null) {
			System.setSecurityManager(new RMISecurityManager());	
		}
		/*
		 * First, set up the initial BLOB->Machine ID mapping
		 */
		for (b_idx = 0; b_idx < nBLOBS; b_idx ++) {
			blob_mcID[b_idx]  = rnd.nextInt(nMACHINES);
			System.out.println("INITIAL: BLOB # " + 
					   b_idx + ": MACHINE: "+ blob_mcID[b_idx]);
			blob_axsFreq[b_idx] = 0;
		}

		/*
		 * Next obtain VoterManager references from
		 * all the machines 
		 */

		try {	
			for (m_idx = 0; m_idx < nMACHINES; m_idx ++) {
				//String m_suffix = "101" + m_idx;
				String m_suffix = "103";	
				v = (VoterManager)Naming.lookup(mPROTOCOL +  ipPREFIX + m_suffix + "/voteMgr");	
				vMRef[m_idx] = v;
				mcID_IP[m_idx] = ipPREFIX + m_idx; 
							
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	
		/*
		 * Now for every voter, pick a machine that the
		 * voter should be on
		 */
		for (v_idx = 0; v_idx < nVOTERS; v_idx ++) {
			m_idx = rnd.nextInt(nMACHINES);
			ht_VtoM.put(new Integer(v_idx), vMRef[m_idx]);
			System.out.println("VOTER: " + v_idx + " :MACHINE: " + m_idx);	
			vtr_mcID[v_idx] = m_idx; 			
		}

		/*
		 * Step 1: Create all the voters
		 */

		ArrayList vList = new ArrayList();
		VoterInfo v_info = null;
		try {
			createVoters(vList, 
				     ht_VtoM, 
				     vtr_mcID, 
				     blob_mcID, 
				     faulty_voters, 
				     mcID_IP);

			v_info = new VoterInfo(vList, VoterConsts.e_UNINIT);
		} catch (Exception e) {
			e.printStackTrace();
		}	
		

		/*
		 * Set the type of network between the machines
	 	 */		

		v_info.setNetworkType(VoterConsts.e_ETHERNET);
		//v_info.setNetworkType(VoterConsts.e_TOKENRING);

		/*
		 * Step 2: Allow all voters to access
		 * their BLOBS independently
		 */


		try {
			startBlobAccess(v_info, VoterConsts.ACCESSTIME);
		} catch (Exception e) {
			e.printStackTrace();
		}

		/*
		 * Step 3: Print the Blob Accesses by
		 * voter. This is a debug feature
	 	 */

		try {
			printBlobFreq(v_info, VoterConsts.e_BASICFREQ);
			printBlobFreq(v_info, VoterConsts.e_OLDDISTFREQ);

		} catch (Exception e) {
			e.printStackTrace();
		}

		/*
	         * Obtain global Blob Access frequencies
		 */		
		try {
			updateBlobAxsFreq(v_info, blob_axsFreq);			

		} catch (Exception e) {
			e.printStackTrace();
		}
		/*
		 * Step 4: Do Byzantine one BLOB after another
		 */
		try {
			for (b_idx = 0; b_idx < nBLOBS; b_idx ++) {
				if (blob_axsFreq[b_idx] < VoterConsts.e_MINBLOBAXSFREQ) {
					System.out.print("SKIPPING BYZ FOR BLOB # " + b_idx);
					System.out.println("LOW AXS FREQ: " + blob_axsFreq[b_idx]);
					continue;	
				}
				System.out.println("STARTING BYZ FOR BLOB # " + b_idx);
				doByzantine(b_idx, v_info);
				System.out.println("BYZ COMPLETE FOR BLOB # " + b_idx);
			}

		} catch (Exception e) {
			e.printStackTrace();
		}
		/*
		 * Step 5: Print the new Blob Distances by Voter
		 */
		try {
			printBlobFreq(v_info, VoterConsts.e_NEWDISTFREQ);	

		} catch (Exception e) {
			e.printStackTrace();
		}


	}
	public static void doByzantine(int blob_id, VoterInfo v)
		throws RemoteException, InterruptedException
	{
		v.setBlob(blob_id);
		v.setMode(VoterConsts.e_DOBYZANTINE);
		v.startVoters();
		System.out.println("STARTING ALL VOTERS FOR BLOB# " + blob_id + "\n");
		while (!v.isDone()){
			System.out.println("MAIN THREAD SLEEPING\n");
			Thread.currentThread().sleep(100);
			;
		}
		System.out.println("STOPPING ALL VOTERS\n");
		v.stopVoters();		
	}

	public static void startBlobAccess(VoterInfo v, long time)
		throws InterruptedException, RemoteException
	{

		v.setMode(VoterConsts.e_AXSBLOB);	
		v.startVoters();
		System.out.println(" MAIN THREAD SLEEPING FOR " + time + " milliseconds");
		Thread.currentThread().sleep(time);
		v.stopVoters();
	}
	public static void printBlobFreq(VoterInfo v, int mode)
		throws RemoteException
	{
		ArrayList vList = v.getVoters();
		for (int idx = 0; idx < vList.size(); idx ++) {
			Voter vtr = ((Voter)vList.get(idx));
			vtr.printBlobFreq(mode);
		}
		

	}

	/*
	 * In this method, you need to compute the relative
	 * access frequencies for all BLOBS in the system
	 */
	public static void updateBlobAxsFreq(VoterInfo v, int[] bFreq)
		throws RemoteException
	{
		ArrayList vList = v.getVoters();
		Hashtable htBlob;
		for (int idx = 0; idx < vList.size(); idx ++) {
			Voter vtr = ((Voter)vList.get(idx));
			htBlob = vtr.getBlobAccessCounts();
			for (Enumeration e = htBlob.keys(); e.hasMoreElements(); ) {
				Integer b_id = (Integer)e.nextElement();
				Integer b_value = ((Integer)htBlob.get(b_id));
				bFreq[b_id.intValue()] += b_value.intValue();	
			}	
		}
		int total_access = 0;
		for (int idx = 0; idx < bFreq.length; idx ++) {
			total_access += bFreq[idx];			
		}

		/* Finally update all elements in bFreq */
		for (int idx = 0; idx < bFreq.length; idx ++) {
			bFreq[idx] *= 100;
			bFreq[idx] /= total_access;
		}		

	}	

}
