import java.awt.event.*;
import java.util.*;
import javax.swing.Timer;

/**
	VoterByzantineImpl Class
	Provides an implementation of a Voter that uses
	Byzantine process to arrive at deciding machines
	to replicate a BLOB.

	@author Preethi Vishwanath
	@version 1.0
*/

public class VoterByzantineImpl extends Timer 
	implements ActionListener, Voter
{

  /**
	Constructor: The VoterManager provides the delay for this voter to sleep
	before casting its own vote, and whether this is a faulty voter or not.
	
	@param:
		delay: Sleep time before voting
		flty: Voter can change vote even in a round.
		mcID: Machine-ID where the voter is located
		mBlobMachine: Hashtable that will contain the eventual blob->Machine mapping
		blob_mcID: Initial Blob->Machine mapping
		mID_IP: Machine-ID to IP Address mapping
	@returns None
    */

	VoterByzantineImpl(int delay, 
			boolean flty,
			int vID,
			int mcID,
		      	Hashtable mblobMachine,
		      	int[] b_mcID,
			String[] mID_IP)
	throws ClassNotFoundException 
	{      
	
		super(delay, null);
		voterFaulty = flty;
		blobMachine = mblobMachine;

		m_id = mcID;
		v_id = vID;
		
		nMachines = mID_IP.length;
		mcID_IP = new String[mID_IP.length];
		for (int idx = 0; idx < mID_IP.length; idx ++) {
			mcID_IP[idx] = mID_IP[idx];
		}
		blob_mcID = new int[b_mcID.length];
		blob_new_mcID = new int[b_mcID.length];
		for (int idx = 0; idx < b_mcID.length; idx ++) {
			blob_mcID[idx] = b_mcID[idx];
			// Initially set it to the same as the old machine
			blob_new_mcID[idx] = b_mcID[idx];
		}
		v_info = null;
		addActionListener(this);
	
	}

	/**
	 * Sets up this voter 
	 * At this point the timer is off,
	 * and the updated VoterInfo is passed
	 * into this object
	 */

	public void setupVoter(VoterInfo vtr_info)
		throws java.rmi.RemoteException
	{
		int m_mode = vtr_info.getMode();
		switch(m_mode) {
			case VoterConsts.e_AXSBLOB:
				v_info = vtr_info;
				initBlobAccess();
				break;
			
			case VoterConsts.e_DOBYZANTINE:
				v_info = vtr_info;
				initByzantine(v_info.getBlob());
				break;			
		}

	}
	/**
	 * Initialize access counts for all Blobs
	 * that this machine is interested in to zero
	 *
	 */  

	private void initBlobAccess()
	{

		blobAxsCnt = new Hashtable();
		blobBasicFreq = new Hashtable();
		for(Enumeration e = blobMachine.keys(); e.hasMoreElements(); ) {
			Integer b_id = (Integer)e.nextElement();
			blobAxsCnt.put(b_id, new Integer(0));
			blobBasicFreq.put(b_id, new Integer(0));
		}

	}
	/**
	 * Updates the access counts for all Blobs
	 * that this machine is interested in.
	 * The access counts are scaled based
	 * on getDistance() function that returns the
	 * distance between two machines, based on a network 
	 * topology. The scaled access counts are obtained thus:
	 * Scaled access[b] = (actual_access[b] * distance[b])
	 * where b is the blob that this voter accesses,
	 * and distance[b] = distance between the machine that this
	 * voter is on, and the machine that the blob resides on.
	 * The program supports two underlying networks connecting the
	 * machines (a) Token ring, and (b) Regular ethernet with all machines
	 * on a single subnet.
	 */

	private void updateBlobAccess()
	{

		Random rVote = new Random();
		int vote = VoterConsts.UNDEFINED;
		try {
			vote = rVote.nextInt(blobAxsCnt.size());
		} catch (Exception e) {
			e.printStackTrace();
			System.err.println("VOTER: " + v_id + " :MC: " + m_id);
			System.exit(1);
		}
		int distance = 0;
		int ht_idx = 0;
		for (Enumeration e = blobAxsCnt.keys(); e.hasMoreElements(); ) {
			Integer key = (Integer)e.nextElement();
			if (vote == ht_idx) {
				int val = ((Integer)blobAxsCnt.get(key)).intValue();
				int basic_freq = ((Integer)blobBasicFreq.get(key)).intValue();
				int b_mc_id = blob_mcID[key.intValue()];
				System.out.print("VOTER: " + v_id + " :MC: " + m_id);
				System.out.print(" :AXS BLOB: " + key.intValue() + " :ON MC: " + b_mc_id);
				distance = getMachineDistance(m_id, b_mc_id);
				System.out.print(" :MC-MC DIST: " + distance + " :OLD CNT: " + val);
				val += distance;
				basic_freq += 1;			
				System.out.println(" :NEW CNT: " + val);
				blobAxsCnt.put(key, new Integer(val));
				blobBasicFreq.put(key, new Integer(basic_freq));		
				break;	
			}
			ht_idx ++;
		}

	}

	/**
	 * Computes the distance between two machines. The distance can be computed
	 * based on 2 mechanisms: Token ring or regular ethernet.
	 * Token ring obtains the distance based on a logical ring
	 * Ethernet obtains the distance assuming a single subnet.
	 * Only the last of the quad is considered for the purpose of determining
	 * the distance. 
	 * This is a utility function. If there are other interesting networks,
	 * one would need to extend this code.
	 * @arguments: Two machine identifiers
	 * @returns: Distance, as a number
	 */
	
	private int getMachineDistance(int src_id, int dst_id)
	{
		int m_distance = 0;
		int src_nib;
		int dst_nib;
		switch (v_info.getNetworkType()) {
			case VoterConsts.e_TOKENRING:
				m_distance = (src_id > dst_id)?(src_id - dst_id):(dst_id - src_id);
				if (m_distance > (nMachines - m_distance)) {
					m_distance = nMachines - m_distance;
				}  	
				break;
			case VoterConsts.e_ETHERNET:
				src_nib = getParsedIP(mcID_IP[src_id], 3);
				dst_nib = getParsedIP(mcID_IP[dst_id], 3);
				m_distance = (src_nib > dst_nib)?(src_nib - dst_nib):(dst_nib - src_nib);
				break;

		}
		return m_distance;		

	}
	/**
	 * Prints the frequency of access to each BLOB by each voter.
	 * Operates in 3 modes:
	 * (a) Basic blob freq of access
	 * (b) Scaled Blob freq of access BEFORE Byzantine
	 * (c) Scaled Blob freq of access AFTER Byzantine
	 * This is purely a debug routine invoked by the manager to display
	 * usage counts for every blob by every voter
	 */
	public void printBlobFreq(int opt)
		throws java.rmi.RemoteException
	{
		
		for (Enumeration e = blobBasicFreq.keys(); e.hasMoreElements(); ) {
			Integer blob_id = (Integer)e.nextElement();
			int old_mc_id = blob_mcID[blob_id.intValue()];
			int new_mc_id = blob_new_mcID[blob_id.intValue()];
			int old_mc_dist = getMachineDistance(m_id, old_mc_id);
			int new_mc_dist = getMachineDistance(m_id, new_mc_id);
			if (new_mc_dist > old_mc_dist) {
				new_mc_dist = old_mc_dist;
			}
			int basic_freq = ((Integer)blobBasicFreq.get(blob_id)).intValue();
			int old_freq = (old_mc_dist)*(basic_freq);
			int new_freq = (new_mc_dist)*(basic_freq);
			
			switch(opt) {
				case VoterConsts.e_BASICFREQ:
					System.out.print("STAT_BASIC_FREQ: VOTER: " + v_id );
					System.out.print(" :MC: " + m_id );
					System.out.print(" :BLOB: " + blob_id );
					System.out.print(" :MC: " + old_mc_id );
					System.out.println(" :FREQ: " + basic_freq );
					break;
				case VoterConsts.e_OLDDISTFREQ:
					System.out.print("STAT_OLD_FREQ: VOTER: " + v_id );
					System.out.print(" :MC: " + m_id );
					System.out.print(" :BLOB: " + blob_id );
					System.out.print(" :MC: " + old_mc_id );
					System.out.print(" :BASE_FREQ: " + basic_freq );
					System.out.println(" :SCALED_FREQ: " + old_freq );
					break;
				case VoterConsts.e_NEWDISTFREQ:
					System.out.print("STAT_NEW_FREQ: VOTER: " + v_id );
					System.out.print(" :MC: " + m_id );
					System.out.print(" :BLOB: " + blob_id );
					System.out.print(" :OLD_MC: " + old_mc_id );
					System.out.print(" :OLD_MC_DIST: " + old_mc_dist );
					System.out.print(" :NEW_MC: " + new_mc_id );
					System.out.print(" :NEW_MC_DIST: " + new_mc_dist );
					System.out.print(" :BASE_FREQ: " + basic_freq );
					System.out.print(" :OLD_SCALED_FREQ: " + old_freq );
					System.out.println(" :NEW_SCALED_FREQ: " + new_freq );
					break;				

			}		

		}
		

	}
	/**
	 * Initializes the Byzantine process for a specific Blob
	 * @arguments: blobID: The Blob for which the Byzantine
	 * 		       process is being initialized
	 */	
	
	private void initByzantine(int blobID) 
		throws java.rmi.RemoteException
	{
      		numVoters = (v_info.getVoters()).size();
		mPos = 0;
		new_pos = 0;
      		isDone = VoterConsts.FALSE;
		agree_m = new StringBuffer(); 
		if (v_info.getNetworkType() == VoterConsts.e_ETHERNET) {
			initByzantine_eth(blobID, mPos, null);
		} else {
     			voteCount = new Vector();
      			m_Votes = new Vector();
      			roundNumber = 0;
			currentWeight = VoterConsts.IDLE_VOTE_WEIGHT;
      			decision = VoterConsts.UNDECIDED;
      			agreement = VoterConsts.UNDECIDED;
			currentVote = setInitialState(blobID);
			Integer n_blob_id = new Integer(blobID);
			if (isIdle == VoterConsts.FALSE &&
				blobAxsCnt.containsKey(n_blob_id)){
				currentWeight = ((Integer)blobAxsCnt.get(n_blob_id)).intValue();
			}
			
		}
	}
	/**
	 * Utility function: Sets the Idle state
	 * based on whether this voter has interest in the 
	 * blob or not
	 */

	private int setInitialState(int blob_id)
	{

    		int tmp_vote;
		if (!blobMachine.containsKey(new Integer(blob_id))) {
			isIdle = VoterConsts.TRUE;
			tmp_vote = (new Random()).nextInt(nMachines);
     		 } else {
			tmp_vote = m_id;
			isIdle = VoterConsts.FALSE;
      		 }
		 System.out.println("VOTER: " + v_id + "MACHINE: " + m_id + "INITIAL VOTE: " + tmp_vote);
		 return tmp_vote;

	}




	/**
	 * This function initializes the vote in a given round
	 * for an ethernet type network
	 * The vote is initialized based on the following factors
	 * If a previous agreement was reached, and the previous agreement
	 * does not correspond with my IP address for that part, then I will
	 * randomly change my vote in favor of a machine that has the previous
	 * agreement set correct
	 * @arguments: blobID: Blob for which this process is taking place
	 * 	     : pos: Position location in the IP Quad for which this
	 * 		    agreeement is taking place
	 * 	     : prev_agreement: Previously arrived agreement. This
	 *			       contains the complete string representation
	 *			       formed from each of the quads for which
	 *			       Byzantine has been performed.
	 *	     : 	mcID: My current vote    
	 */
	

	public int initByzantine_eth(int blobID, int pos, String prev_agreement)
		throws java.rmi.RemoteException
	{
		String chosen_mc_ip;
		String[] tmp_result;
		StringBuffer str_result;
		int rVal = VoterConsts.UNDEFINED;
		int idx = 0;
		int p_idx = 0;
		partial_mcs = new Vector();	
		boolean my_mc_present = false;
		currentWeight = VoterConsts.IDLE_VOTE_WEIGHT;
      		roundNumber = 0;
      		decision = VoterConsts.UNDECIDED;
      		agreement = VoterConsts.UNDECIDED;
	
     		voteCount = new Vector();
      		m_Votes = new Vector();
		mPos = pos;

		int mcID = setInitialState(blobID);		
		for (idx = 0; idx < nMachines; idx ++) {
			if (prev_agreement == null) {
				partial_mcs.add(new Integer(idx));
				my_mc_present = true;
				continue;
			}
			tmp_result = (mcID_IP[idx]).split("\\.");
			str_result = new StringBuffer();
			p_idx = 0;
			while (p_idx < pos) {
				str_result.append(tmp_result[p_idx]).append(".");		
				p_idx ++;
			}
			if ((str_result.toString()).equalsIgnoreCase(prev_agreement)) {
				partial_mcs.add(new Integer(idx));
				if (idx == mcID) {
					my_mc_present = true;
				}
			}	

		}
		if (my_mc_present) {
			chosen_mc_ip = mcID_IP[mcID];
			Integer n_blob_id = new Integer(blobID);
			if (blobAxsCnt.containsKey(n_blob_id)){
				currentWeight = ((Integer)blobAxsCnt.get(n_blob_id)).intValue();
			}
		} else {
			// Pick at random any of the machines
			// and return the quad at that point
			// You have effectively turned into an Idle voter

			isIdle = VoterConsts.TRUE;
			idx = new Random().nextInt(partial_mcs.size());
			p_idx = ((Integer)partial_mcs.elementAt(idx)).intValue();
			chosen_mc_ip = mcID_IP[p_idx];		
		}
		rVal = getParsedIP(chosen_mc_ip, pos);
		System.out.println("VOTER: " + v_id + " POSITION " + mPos + " CURRENT VOTE " + rVal);	
		currentVote = rVal;
		return rVal;
	}

   /**

   broadcastRandom: Faulty Voters will call this function to send random
                    votes to each of the other Voters in the electorate.
                    Each round all of the vote values that were sent to 
                    the other Voters are output to the console.

   @return void

   */
   
   public void broadcastRandom() 
	throws java.rmi.RemoteException 
 {
	Random voteSelector = new Random();
	if (v_info.getNetworkType() == VoterConsts.e_ETHERNET) {
		int mc_idx = 0;
		int mc_voted = VoterConsts.UNDEFINED;
		mc_idx = voteSelector.nextInt(partial_mcs.size());
		mc_voted = ((Integer)partial_mcs.elementAt(mc_idx)).intValue();
		currentVote = getParsedIP(mcID_IP[mc_voted], mPos);					
	} else {
		currentVote = voteSelector.nextInt(nMachines);
	}
	broadCast(currentVote, currentWeight);
  
}
   /**

   broadcastNormal: Normal Voters will call this function to send the same
                    vote to all of the other Voters in the electorate.
                    Each round all of the vote value that was sent to 
                    the other Voters is output to the console.
		    This voter does not change his vote in the same round

   @return void

   */
   
   public void broadcastNormal()  
	throws java.rmi.RemoteException 
   {
	
	broadCast(currentVote, currentWeight);	

   }

   private void broadCast(int Vote, int times)  
	throws java.rmi.RemoteException 
  {
	ArrayList electorate = v_info.getVoters();	
	// Broadcast this vote to everyone
	for(int i = 0; i < numVoters; i++) {
		if (i == v_id) continue;
     		((Voter)electorate.get(i)).receiveVote(Vote, times, this);
      	}


  }


/**
 * Utility function that returns the parsed IP quad, as a function of the position
 * 
 * @arguments: ip_addr: String representing the IP Address in dotted decimal
 *	     : pos: Position in the dotted decimal from which to return the value
 * @returns: The dotted decimal value at pos
 */

private int getParsedIP(String ip_addr, int pos)
{
	String[] result = ip_addr.split("\\.");
	int rval = Integer.parseInt(result[pos]);
	return rval;	
} 

// This hashtable keeps a track of the votes
// for every machine. Since we might have non-contiguous
// numbers in the ethernet case,.
// AND the votes getting passed around are quads, this
// makes a whole lot of sense.

private Hashtable initializeVoteTable()
{
	Hashtable ht_VoteM = new Hashtable();
	for (int idx = 0; idx < nMachines; idx ++) {
		Integer curr_quad = null;
		if (v_info.getNetworkType() == VoterConsts.e_ETHERNET) {
			curr_quad = new Integer(getParsedIP(mcID_IP[idx], mPos));
		} else {
			curr_quad = new Integer(idx);
		}
		if (!ht_VoteM.containsKey(curr_quad)) {
			ht_VoteM.put(curr_quad, new Integer(0));
		}		
	}
	return ht_VoteM;

}
 /**

   receiveVote: Maintains the running count of the votes received from 
                all of the other Voters for every round executed

   @param m_idx  What the vote is (one of the machineIDs, either the quad or the actual ID)
   @param v  Which Voter cast that vote
   
   @return void  

   */     
   
   public void receiveVote (int m_idx, int weight, Voter v) 
	throws java.rmi.RemoteException 

   {
      int round = v.getRoundNumber();

      //voteCount vector keeps a running count of all
      //votes received in this round	
     // For every round, there is one entry that is updated
     
      if (voteCount.size() < (round + 1)) {
		voteCount.add(new Integer(0));
		m_Votes.add(initializeVoteTable());	
      }
      int voteVal = ((Integer)voteCount.get(round)).intValue();	
      voteCount.set(round, new Integer(voteVal + 1));
      Hashtable ht_mach = (Hashtable)m_Votes.get(round);
      Integer mc_Key = new Integer(m_idx);
      int curr_votes = ((Integer)ht_mach.get(mc_Key)).intValue();
      ht_mach.put(mc_Key, new Integer(curr_votes + weight)); 	     
 
   }

/**

   actionPerformed: This method implements different actions based on
		    the phase that the object currently in.
		    The phase that the object is currently in
		    is determined by the mode setting set previously        
   @param arg0  ActionEvent is not used

   @return void

*/

public void actionPerformed(ActionEvent arg0) 
{
	int mode = v_info.getMode();
	switch(mode) {
		case VoterConsts.e_AXSBLOB:
			updateBlobAccess();
			break;
		case VoterConsts.e_DOBYZANTINE:
			try {
				doByzantine();
			} catch (Exception e) { e.printStackTrace(); }
			break;
	}
}
/**
 * The main loop for performing Byzantine agreement
 * Bascially, this function is split into whether we're doing Byzantine on
 * a token ring or an ethernet type network.
 * For token ring, there is only one Byzantine. 
 * For ethernet, you need to perform Byzantine until all quads have
 * been completed
 */

private void doByzantine()
	throws java.rmi.RemoteException
{
	if (v_info.getNetworkType() == VoterConsts.e_ETHERNET) {
		doByz_eth();
	} else {
		doByz_other();
	}
}


private void doByz_eth()
	throws java.rmi.RemoteException
{

	ArrayList electorate = v_info.getVoters();
	if (new_pos == VoterConsts.IP_LENGTH) {
		for (int idx = 0; idx < numVoters; idx ++) {
			Voter v = (Voter)(electorate.get(idx));
			v.setDone(VoterConsts.TRUE);
		}
		System.out.print("*****BYZANTINE AGREEMENT REACHED FOR BLOB*****" + v_info.getBlob());
		System.out.println("VOTER ID " + v_id + " AGREEMENT = " + agree_m.toString());
		stop();
		return;
	}

	if ((new_pos > mPos) && (v_id == 0)) {
		int count = 0;
		for (int idx = 0; idx < numVoters; idx ++) {
			Voter v = (Voter)(electorate.get(idx));
			if (v.getPos() == new_pos) {
				count += 1;
			}
		}
		if (count == numVoters) {
			// All have reached the same point
			// First stop myself
			System.out.println("STOPPING LEADER VOTER #:" + v_id);
			stop();
			mPos = new_pos;
			for (int idx = 0; idx < numVoters; idx ++) {
				Voter v = (Voter)electorate.get(idx);
				v.initByzantine_eth(v_info.getBlob(), mPos, agree_m.toString());  	
			}
			
			for (int idx = 0; idx < numVoters; idx ++) {
				Voter v = (Voter)electorate.get(idx);
				v.start();  	
			}

		}
		return;			
	}
	
	if (doByz()) {
		agree_m.append(agreement);
		if (mPos < (VoterConsts.IP_LENGTH - 1)) {
			agree_m.append(".");
		}
		System.out.print("VOTER: " + v_id + " :BLOB: " + v_info.getBlob());
		System.out.println("POSITION: " + mPos + " :AGREEMENT IS: " + agree_m.toString());
		if (mPos == (VoterConsts.IP_LENGTH - 1)) {
			String[] final_result = agree_m.toString().split("\\.");
			blob_new_mcID[v_info.getBlob()] = Integer.parseInt(final_result[mPos]);
							
		}
		new_pos = mPos + 1;
		if (v_id != 0)  {
			System.out.println("STOPPING VOTER #" + v_id);
			stop();
		}
	}
}

private void doByz_other()
	throws java.rmi.RemoteException
{
	if (doByz()) {
		isDone = VoterConsts.TRUE;
		agree_m.append(agreement);
		System.out.println("***AGREEMENT REACHED****" + agree_m.toString());
		blob_new_mcID[v_info.getBlob()] = agreement;
		stop();
	}	
}

private boolean doByz()
	throws java.rmi.RemoteException
{
	if (roundNumber == 0) {  
		voteByzantine();
   	} else {
     		if (((Integer)voteCount.get(roundNumber - 1)).intValue() < (numVoters - 1)) {
     			return false;
		}
     		if (agreementReached()) {
			voteByzantine();
			System.out.println("Voter " + v_id + 
				"  Agreement Reached!!"
                                + " Agreement is " + agreement);
			return true;
	
		} else {
			voteByzantine();
		}
	}
	return false;
}

/*
   voteByzantine: Implements the Byzantine election algorithm. Checks 
   for a majority vote and determines if the tally of the
   majority vote exceeds the threshold or not.  The threshold
   is determined by a global coin toss.  If the tally exceeds
   the threshold then the Voter votes the same as the majority, 
   otherwise it votes 0.  When 7/8 of all Voters reach the same 
   decision then that decision becomes permanent for this voter.
                  
   @return void

*/
   
public void voteByzantine () 
	throws java.rmi.RemoteException 
{
	if (isIdle || voterFaulty) {
		broadcastRandom();
	} else {
		broadcastNormal();
	}
	if (voteCount.size() < (roundNumber + 1)) {
		voteCount.add(new Integer(0));
		m_Votes.add(initializeVoteTable());	
		
	}
        if (roundNumber != 0) {
		Hashtable ht_voteM = (Hashtable)(m_Votes.get(roundNumber - 1));
		Integer currVKey = new Integer(currentVote);
		Integer myVoteCnt = (Integer)ht_voteM.get(currVKey);
		ht_voteM.put(currVKey, new Integer(myVoteCnt.intValue() + currentWeight));	
	 	System.out.print("VTR[" + v_id + "]:VOTE:[" + currentVote + "]:RND[" + roundNumber + "]"); 
	 	
		int maj_machine = VoterConsts.UNDEFINED;
	 	int maj_machine_tally = VoterConsts.UNDEFINED;
		
		int total_votes = 0;
		int mach_vote_val = 0;
	 	for (Enumeration e = ht_voteM.keys(); e.hasMoreElements(); ) {
			Integer mach_id = (Integer)e.nextElement();				
			mach_vote_val = ((Integer)ht_voteM.get(mach_id)).intValue();			

			if (mach_vote_val > maj_machine_tally) {
				maj_machine_tally = mach_vote_val;
				maj_machine = mach_id.intValue();
			}
			total_votes += mach_vote_val;	
	 	}
	
	 	int threshold;
         	threshold = (int) (((5.0 / 8.0) * total_votes) + 1);
         	currentVote = (maj_machine_tally >= threshold) ? maj_machine : currentVote;
			
		/*	
		// threshold for changing idle is much lower
		int nflty = 0;
		int nidle = 0;	
		ArrayList electorate = v_info.getVoters();		

		for (int i = 0; i < numVoters; i ++) {
			if (((Voter)electorate.get(i)).isFaulty()) {
				nflty += 1;
				continue;
			}
			if (((Voter)electorate.get(i)).isIdle()) {
				nidle += 1;
			}
		
		}	
		int numNormal = numVoters - nflty - nidle;
	
		if ((maj_machine_tally >= ((3.0 / 4.0) * numNormal)) && isIdle) {
			isIdle = false;
			currentVote = maj_machine;			
		} 
		*/
	
         	if ((maj_machine_tally >= ((7.0 / 8.0) * mach_vote_val)) 
				&& (decision == VoterConsts.UNDECIDED)) {
	   		 decision = maj_machine;
			 decisionVotes = maj_machine_tally;
            		 System.out.println( "ROUND: " + roundNumber + " VOTER: " + v_id + " DECISION: " + decision);
		}
      	}
      	roundNumber++; 
   }

   /**
	isDone: Query function for the top level manager, basically
	a status check to see if this node is done or NOT
	
	@param none
	@return boolean
    */
  
   public void setDone(boolean done_val)
	throws java.rmi.RemoteException
   {
	isDone = done_val;
   }
   public boolean isDone() 
	throws java.rmi.RemoteException
   {
	return (isDone == VoterConsts.TRUE);
   }
   public boolean isFaulty()
	throws java.rmi.RemoteException 
   {
	return voterFaulty;
   } 	
   public boolean isIdle() 
	throws java.rmi.RemoteException
   {
	return (isIdle == VoterConsts.TRUE);
   }
   /**

   getDecision: gets a Voter's decision

   @return int 

   */   
   
   public int getDecision()
   {
      return decision;
   }

   public int getDecisionVotes()
   {
      return decisionVotes;
   }
   
   /**

   getRoundNumber: gets the round number that a Voter is currently in 

   @return int  round number

   */
   
   public int getRoundNumber()
   {
      return roundNumber;
   }
  
   public int getPos()
   {
	return new_pos;
   } 
   
 /**

   agreementReached: Checks if 7/8 of all the Voters have come to a final
                     decision or not
   
   @return boolean  1 if agreement has been reached 0 if not 

   */  
   
   public boolean agreementReached () 
	throws java.rmi.RemoteException
   {
      boolean result = false;
      Hashtable nm_count = new Hashtable();
      
      int mach_vote = VoterConsts.UNDEFINED;
      int mach_vote_val = VoterConsts.UNDEFINED;
		
      ArrayList electorate = v_info.getVoters();
 
      for(int i = 0; i < numVoters; i++)  {
	  mach_vote = ((Voter)electorate.get(i)).getDecision();
	  mach_vote_val = ((Voter)electorate.get(i)).getDecisionVotes();
	  if (mach_vote == VoterConsts.UNDECIDED) continue;
	  Integer mach_ID = new Integer(mach_vote);
	  if (!nm_count.containsKey(mach_ID)) {
		nm_count.put(mach_ID, new Integer(mach_vote_val));
	  } else {
		Integer mch_val = (Integer)nm_count.get(mach_ID);
		nm_count.put(mach_ID, new Integer(mch_val.intValue() + mach_vote_val));
	  }
      }
      int maj_val = VoterConsts.UNDEFINED;
      int maj_val_machine = VoterConsts.UNDEFINED;
      int total_votes = 0;
      for (Enumeration e = nm_count.keys(); e.hasMoreElements(); ) {
		Integer mch_key = (Integer)e.nextElement();
		Integer mch_value = (Integer)nm_count.get(mch_key);
		if (mch_value.intValue() > maj_val) {
			maj_val = mch_value.intValue();
			maj_val_machine = mch_key.intValue();
		}
		total_votes += mch_value.intValue();		
      }
  
      if (maj_val >= ((7.0 / 8.0) * total_votes)) {
	 agreement = maj_val_machine;
         result = true;
      }
      
      return result;   
   }
   public Hashtable getBlobAccessCounts()
		throws java.rmi.RemoteException
   {
	Hashtable blobScaledFreq = new Hashtable();
	for (Enumeration e = blobBasicFreq.keys(); e.hasMoreElements(); ) {
		Integer blob_id = (Integer)e.nextElement();
		/*
		 * Since the array is updated after performing Byzantine
		 * calling this method at different points, provides
		 * different access frequencies for different BLOBS
		 */

		int mc_id = blob_new_mcID[blob_id.intValue()];
		int mc_distance = getMachineDistance(m_id, mc_id);
		int basic_freq = ((Integer)blobBasicFreq.get(blob_id)).intValue();
		blobScaledFreq.put(blob_id, new Integer(mc_distance * basic_freq));
	}
	return blobScaledFreq;
     }

   private boolean voterFaulty; // Idle,faulty or not?
   private boolean isDone; // Am I done, and in which round?
			     // The number of bits on which we have
			     // to agree is passed 
   private String[] mcID_IP;
   int m_id; // Machine ID for this voter
   private int roundNumber; // current round
   private int v_id;
   private int mPos;
   private int new_pos;
   private int currentVote; // Voter's vote for current round
   private int currentWeight; // Voter's weight for his vote

   private int decision; // final decision for this Voter
   private int decisionVotes;
   private int agreement; // agreement of at least 7/8 of Voters
   private StringBuffer agree_m;
   private int numVoters; // Total number of Voters
   
   private boolean isIdle;
   private int nMachines; 
   Hashtable blobMachine;
   Hashtable blobAxsCnt;
   Hashtable blobBasicFreq;

   private VoterInfo v_info;

  // Keeps track of the number of votes
  // per machine
   private Vector m_Votes;
   private Vector voteCount;
   private Vector partial_mcs;	

   private static final long serialVersionUID = 1L;
   private int[] blob_mcID;
   private int[] blob_new_mcID;
}
